package ctbrec.recorder.postprocessing;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.util.Arrays;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ctbrec.OS;
import ctbrec.Recording;
import ctbrec.io.StreamRedirectThread;
import ctbrec.recorder.download.ProcessExitedUncleanException;

public class Remuxer extends AbstractPostProcessor {

    private static final Logger LOG = LoggerFactory.getLogger(Remuxer.class);

    public static final String FFMPEG_ARGS =  "ffmpeg.args";
    public static final String FILE_EXT =  "file.ext";

    @Override
    public String getName() {
        return "remux / transcode";
    }

    @Override
    public void postprocess(Recording rec) throws IOException, InterruptedException {
        String fileExt = getConfig().get(FILE_EXT);
        String[] args = getConfig().get(FFMPEG_ARGS).split(" ");
        String[] argsPlusFile = new String[args.length + 3];
        int i = 0;
        argsPlusFile[i++] = "-i";
        argsPlusFile[i++] = rec.getPostProcessedFile().getAbsolutePath();
        System.arraycopy(args, 0, argsPlusFile, i, args.length);
        File remuxedFile = new File(rec.getPostProcessedFile().getAbsolutePath() + '.' + fileExt);
        argsPlusFile[argsPlusFile.length - 1] = remuxedFile.getAbsolutePath();
        String[] cmdline = OS.getFFmpegCommand(argsPlusFile);
        LOG.info(Arrays.toString(cmdline));
        Process ffmpeg = Runtime.getRuntime().exec(cmdline, new String[0], rec.getPostProcessedFile().getParentFile());
        setupLogging(ffmpeg, rec);
        rec.setPostProcessedFile(remuxedFile);
        rec.getAssociatedFiles().add(remuxedFile.getAbsolutePath());
    }

    private void setupLogging(Process ffmpeg, Recording rec) throws IOException, InterruptedException {
        int exitCode = 1;
        File video = rec.getPostProcessedFile();
        File ffmpegLog = new File(video.getParentFile(), video.getName() + ".ffmpeg.log");
        try (FileOutputStream mergeLogStream = new FileOutputStream(ffmpegLog)) {
            Thread stdout = new Thread(new StreamRedirectThread(ffmpeg.getInputStream(), mergeLogStream));
            Thread stderr = new Thread(new StreamRedirectThread(ffmpeg.getErrorStream(), mergeLogStream));
            stdout.start();
            stderr.start();
            exitCode = ffmpeg.waitFor();
            LOG.debug("FFmpeg exited with code {}", exitCode);
            stdout.join();
            stderr.join();
            mergeLogStream.flush();
        }
        if (exitCode != 1) {
            if (ffmpegLog.exists()) {
                Files.delete(ffmpegLog.toPath());
            }
        } else {
            rec.getAssociatedFiles().add(ffmpegLog.getAbsolutePath());
            LOG.info("FFmpeg exit code was {}. Logfile: {}", exitCode, ffmpegLog.getAbsolutePath());
            throw new ProcessExitedUncleanException("FFmpeg exit code was " + exitCode);
        }
    }

    @Override
    public String toString() {
        String s = getName();
        if(getConfig().containsKey(FFMPEG_ARGS)) {
            s += " [" + getConfig().get(FFMPEG_ARGS) + ']';
        }
        return s;
    }
}
